/**
 * Constants
 */
var gc = gc || {};

var numPending = 0;
var maxNumPending = 0;

var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

const Cmd_SPI_Control = 0x08;

var getDeviceIndexSpi = function(coreIndex, regModelId) {
  coreIndex = coreIndex === undefined ? 0 : coreIndex;

  if (coreIndex > -1 && regModelId === 'bq7971x') {
    coreIndex = coreIndex + 1;
  }

  return coreIndex;
};

var init = function() {
  var regReadDetail = undefined,
    regReadCount,
    regReadPromiseArr;

  /**
   * This code implements an example custom USB2ANY Interface (functionally just a clone of the default I2C interface)
   *
   * @constructor
   * @param settings - The interface settings defined in the system.json
   * @param registerModel - The register model
   * @implements gc.databind.USB2ANY.ICommInterface
   */
  CustomSPI_Interface = function(settings, registerModel) {
    this._registerModel = registerModel;
    this.transferPromises = [];
    this.cmdQueue = [];
    this.readyToProcess = true;
    this.noOfDevices = 0;
    this.config = {};
    this.is600 = false;
    this.opPromise;
  };

  /**
   * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
   * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
   * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
   */
  gc.databind.USB2ANY.prototype.CUSTOMSPI_Interface = CustomSPI_Interface;

  /**
   * Custom interfaces should be extended from the ICommInterface base class
   */
  CustomSPI_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

  /**
   * Set the communication parameters
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param settings - The interface settings defined in the system.json
   */
  CustomSPI_Interface.prototype.control = function(u2a, settings) {
    console.log('SPI', 'Control', u2a, settings);

    var clkPhase = settings.clkPhase === 'leadingEdge' ? 1 : 0;
    var clkPolarity = settings.clkPolarity === 'inactiveHigh' ? 1 : 0;
    var bitDir = settings.bitDir === 'MSBFirst' ? 1 : 0;
    var charLen = settings.charLen === 7 ? 1 : 0;
    var csType;
    switch (settings.csType) {
      case 'everyByte':
        csType = 0;
        break;
      case 'everyPacket':
        csType = 1;
        break;
      case 'everyWord':
        csType = 2;
        break;
      case 'noCS':
        csType = 3;
        break;
      case 'everyPacket_MOSIInactiveHigh':
        csType = 4;
        break;
      case 'pulseAfterPacket':
        csType = 255;
        break;
      default:
        csType = 0;
        break;
    }
    var csPolarity = settings.csPolarity === 'activeLow' ? 1 : 0;

    var divHigh = parseInt(settings.divHigh);
    if (isNaN(divHigh)) {
      divHigh = 0;
    } else {
      if (divHigh < 0) {
        divHigh = 0;
      } else if (divHigh > 255) {
        divHigh = 255;
      }
    }

    var divLow = parseInt(settings.divLow);
    if (isNaN(divLow)) {
      divLow = 0;
    } else {
      if (divLow < 0) {
        divLow = 0;
      } else if (divLow > 255) {
        divLow = 255;
      }
    }

    if (divHigh === 0 && divLow === 0) {
      divLow = 4;
    }

    this.config.clkPhase = clkPhase;
    this.config.clkPolarity = clkPolarity;
    this.config.bitDir = bitDir;
    this.config.charLen = charLen;
    this.config.csType = csType;
    this.config.csPolarity = csPolarity;
    this.config.divHigh = divHigh;
    this.config.divLow = divLow;
  };

  CustomSPI_Interface.prototype.configure = function(u2a) {
    this.enable(u2a);

    u2a.onPayloadPacketReceived = this.onPayloadPacketReceived;
    u2a.shouldPauseTransmission = this.shouldPauseTransmission;

    for (var i = 0; i < this.transferPromises.length; i++) {
      this.transferPromises[i].reject();
    }
    this.cmdQueue.splice(0, this.cmdQueue.length);
    this.readyToProcess = true;
  };

  CustomSPI_Interface.prototype.enable = function(u2a) {
    console.log('SPI', 'Config', 'Params', [
      this.config.clkPhase,
      this.config.clkPolarity,
      this.config.bitDir,
      this.config.charLen,
      this.config.csType,
      this.config.csPolarity,
      this.config.divHigh,
      this.config.divLow
    ]);
    u2a.sendCommandPacket(Cmd_SPI_Control, [
      this.config.clkPhase,
      this.config.clkPolarity,
      this.config.bitDir,
      this.config.charLen,
      this.config.csType,
      this.config.csPolarity,
      this.config.divHigh,
      this.config.divLow
    ]);
  };

  CustomSPI_Interface.prototype.disable = function(u2a) {};

  CustomSPI_Interface.prototype.calculateWriteData = function(writeData) {
    var bytesToRead = 0,
      bytesToReadHigh,
      bytesToReadLow,
      waitTime = 0,
      waitTimeHigh,
      waitTimeLow,
      devAddr;
    var initByte = writeData[0];
    var noOfDevices = this.noOfDevices;

    if (initByte === 0x80) {
      devAddr = writeData[1];
      bytesToRead = writeData[4] + 7;
      // if(writeData[1] === 0x00) {
      //     waitTime = 100 + (bytesToRead * 10);
      // } else {
      //     waitTime = ((devAddr - 1) * 6) + (bytesToRead * 10) + 100;
      // }
    } else if (initByte === 0xa0) {
      bytesToRead =
        (writeData[3] + 7) * (noOfDevices > 0 ? noOfDevices - 1 : 0);
      // waitTime = ((noOfDevices > 2 ? noOfDevices - 2 : 0) * 6) + (bytesToRead * 10) + 100;
    } else if (initByte === 0xc0) {
      bytesToRead = (writeData[3] + 7) * noOfDevices;
      // waitTime = ((noOfDevices - 1) * 6) + (bytesToRead * 10) + 100;
    }

    bytesToReadHigh = Math.floor(bytesToRead / 256);
    bytesToReadLow = bytesToRead % 256;

    waitTime = 10000;
    waitTimeHigh = Math.floor(waitTime / 256);
    waitTimeLow = waitTime % 256;

    writeData.unshift(writeData.length);
    writeData.push(bytesToReadHigh);
    writeData.push(bytesToReadLow);
    writeData.push(waitTimeHigh);
    writeData.push(waitTimeLow);

    return writeData;
  };

  CustomSPI_Interface.prototype.shouldPauseTransmission = function(txPacket) {
    return false;
  };

  CustomSPI_Interface.prototype.onPayloadPacketReceived = function(
    target,
    cmd,
    rawDataSliced,
    rawData,
    payloadLen,
    payloadStart
  ) {
    // console.log('SPI', 'onPayloadPacketReceived', target, cmd, rawData);
    if (cmd === 144 || cmd === 146) {
      var transferPromises = this.getInterfaces()['customspi'].transferPromises;
      // console.log('SPI', 'Transfer data in buffer', transferPromises.length);
      if (transferPromises && transferPromises.length > 0) {
        // console.log('SPI', 'Transfer promise resolve');
        var transferPromise = transferPromises.shift();
        // console.log('SPI', 'Transfer promise shifted', transferPromises.length);
        transferPromise.resolve({
          rawData: rawData,
          payloadLen: payloadLen,
          payloadStart: payloadStart
        });
      }
    } else if (cmd === 148) {
      // console.log('SPI', 'onPayloadPacketReceived', target, cmd, rawData, payloadLen);
      var status = rawData.slice(
        payloadStart,
        payloadStart + rawData[payloadLen]
      )[0];
      if (status > 0) {
        // console.log('SPI', 'onPayloadPacketReceived', 'Timeout');
        // this.sendCommandPacket(148, this.getInterfaces()['customspi'].calculateWriteData([0x00]));
        this.sendCommandPacket(9, [1, 0x00]);
      }
      var spiInterface = this.getInterfaces()['customspi'];
      if (spiInterface.opPromise) {
        // console.log('SPI', 'onPayloadPacketReceived', 'Resolve');
        spiInterface.opPromise.resolve();
        spiInterface.opPromise = undefined;
      }
    } else {
      gc.console.debug('USB2ANY', 'Ignoring Payload Packet for command #', cmd);
    }
  };

  CustomSPI_Interface.prototype.recursivePing = function(u2a, count) {
    var deferred = Q.defer();

    // Clear Rx Buffer
    u2a.sendCommandPacket(147, []);

    // 606
    // var writeData = [0x80, count, 0x01, 0x04, 0x00];

    // 616
    var writeData = [0x80, count, 0x03, 0x06, 0x00];

    var crc = getByteArrFromHexStr(
      prependZero(getCRC16IBM(writeData).toString(16), 4)
    );
    writeData.push.apply(writeData, crc);
    writeData = this.calculateWriteData(writeData);

    var opPromise = Q.defer();
    this.opPromise = opPromise;

    opPromise.promise.then(
      function() {
        var buffLenDeferred = Q.defer();
        this.transferPromises.push(buffLenDeferred);
        buffLenDeferred.promise.then(
          function(data) {
            var rawData = data.rawData;
            var payloadLen = data.payloadLen;
            var payloadStart = data.payloadStart;
            var payload = rawData.slice(
              payloadStart,
              payloadStart + rawData[payloadLen]
            );
            var dataLen = payload[1] * 256 + payload[0];

            if (dataLen > 0) {
              this.recursivePing(u2a, count + 1).then(
                function(countData) {
                  deferred.resolve(countData);
                }.bind(this)
              );
            } else {
              deferred.resolve(count);
            }
          }.bind(this)
        );
        u2a.sendCommandPacket(146, []);
      }.bind(this)
    );

    u2a.sendCommandPacket(148, writeData);

    return deferred.promise;
  };

  CustomSPI_Interface.prototype.recursiveRead = function(
    u2a,
    cmdArr,
    deferred,
    readData
  ) {
    var cmd = cmdArr.shift();
    logger('Read', cmd);
    var noOfRegs = (cmd[0] < 160 ? cmd[4] : cmd[3]) + 1;
    var noOfDevices = cmd[0] < 160 ? 1 : this.noOfDevices;

    cmd = this.calculateWriteData(cmd);
    // console.log('SPI', 'Recursive read', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

    var opPromise = Q.defer();
    this.opPromise = opPromise;

    opPromise.promise.then(
      function() {
        if (cmdArr.length > 0) {
          this.recursiveRead(u2a, cmdArr, deferred, readData);
        } else {
          var bufferLenDeferred = Q.defer();
          this.transferPromises.push(bufferLenDeferred);
          bufferLenDeferred.promise.then(
            function(data) {
              // console.log('SPI', 'Recursive read', 'Buffer data length', data);
              var rawData = data.rawData;
              var payloadLen = data.payloadLen;
              var payloadStart = data.payloadStart;
              var payload = rawData.slice(
                payloadStart,
                payloadStart + rawData[payloadLen]
              );

              var dataLen = payload[1] * 256 + payload[0];
              var packetSize = Math.ceil(dataLen / 54);

              function packetsTransferResolve(packet) {
                // console.log('SPI', 'Recursive read', 'Packet', packet);
                var packetPayloadLen = packet.rawData[packet.payloadLen];
                var packetPayload = packet.rawData.slice(
                  packet.payloadStart,
                  packet.payloadStart + packetPayloadLen
                );
                readData.push.apply(readData, packetPayload);
                // console.log('SPI', 'Recursive read', 'Res data', readData);
              }

              var packetDeferred;
              var packetsTransferPromise = [];
              for (var i = 0; i < packetSize; i++) {
                packetDeferred = Q.defer();
                packetsTransferPromise[i] = packetDeferred.promise;
                this.transferPromises.push(packetDeferred);
                packetsTransferPromise[i].then(packetsTransferResolve);
              }
              // console.log('SPI', 'Recursive read', 'Packet transfer promises', packetsTransferPromise);
              Promise.all(packetsTransferPromise).then(
                function() {
                  // console.log('SPI', 'Recursive read', 'Final res', readData);
                  logger('Response', readData);
                  deferred.resolve(readData);
                }.bind(this)
              );

              if (packetSize > 0) {
                // console.log('SPI', 'Recursive read', 'Buffer data transfer', (new Date()).getTime());
                u2a.sendCommandPacket(144, [payload[1], payload[0]]);
              }
            }.bind(this)
          );
          // console.log('SPI', 'Recursive read', 'Get buffer size', 'Timestamp', (new Date()).getTime());
          u2a.sendCommandPacket(146, []);
        }
      }.bind(this)
    );

    u2a.sendCommandPacket(148, cmd);
  };

  CustomSPI_Interface.prototype.manageReads = function(u2a, cmdArr) {
    var deferred = Q.defer();
    var readData = [];

    // Clear Rx Buffer
    u2a.sendCommandPacket(147, []);

    if (cmdArr.length > 0) {
      this.recursiveRead(u2a, cmdArr, deferred, readData);
    } else {
      deferred.resolve(readData);
    }
    return deferred.promise;
  };

  CustomSPI_Interface.prototype.recursiveWrite = function(
    u2a,
    cmdArr,
    deferred
  ) {
    var cmd = cmdArr.shift();
    logger('Write', cmd);
    // console.log('SPI', 'Recursive write', 'Timestamp', (new Date()).getTime());
    cmd = this.calculateWriteData(cmd);
    console.log('SPI', 'Recursive write', cmd);
    // console.log('SPI', 'Custom interface', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

    var opPromise = Q.defer();
    this.opPromise = opPromise;

    opPromise.promise.then(
      function() {
        if (cmdArr.length > 0) {
          this.recursiveWrite(u2a, cmdArr, deferred);
        } else {
          deferred.resolve();
        }
      }.bind(this)
    );

    u2a.sendCommandPacket(148, cmd);
  };

  CustomSPI_Interface.prototype.manageWrites = function(u2a, cmdArr) {
    var deferred = Q.defer();
    if (cmdArr.length > 0) {
      this.recursiveWrite(u2a, cmdArr, deferred);
    } else {
      deferred.resolve();
    }
    return deferred.promise;
  };

  CustomSPI_Interface.prototype.manageReadWrite = function(u2a, cmdInfo) {
    manageReadWriteUtil.bind(this)(u2a, cmdInfo);
  };

  /**
   * Construct and send a command packet for reading from a register
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param info - The register map info for the register
   * @return - A promise that is resolved once the data is received
   */
  CustomSPI_Interface.prototype.read = function(
    u2a,
    info,
    regModel,
    coreIndex
  ) {
    console.log(
      'SPI',
      'Read',
      u2a,
      info,
      regModel,
      coreIndex,
      getDeviceIndexSpi(coreIndex, regModel._id),
      regReadDetail
    );
    var regAddr = info.addr;
    var readData = [],
      crc;

    var deviceIndex = getDeviceIndexSpi(coreIndex, regModel._id);

    var regPage = document.querySelector('ti-widget-register-page');

    if (regReadDetail) {
      // if(regReadDetail && ((regReadDetail.regModelId === regModel._id) && ((coreIndex === undefined) || (regReadDetail.coreIndex === coreIndex)))) {

      if (deviceIndex < 0) {
        var deferred = Q.defer();

        regReadPromiseArr.push({
          promise: deferred,
          addr: info.addr
        });

        regReadCount++;

        if (regReadCount === regReadDetail.numRegs) {
          this.manageReadWrite(u2a, {
            type:
              this.is600 === true
                ? regModel._id === 'bq79600'
                  ? 'readAllBroadcast600'
                  : 'readAllStack'
                : 'readAllBroadcast',
            device: deviceIndex,
            count: regReadCount,
            promiseArr: regReadPromiseArr
          });
        }

        return deferred.promise;
      } else {
        var deferred = Q.defer();

        regReadPromiseArr.push({
          promise: deferred,
          addr: info.addr
        });

        regReadCount++;

        if (regReadCount === regReadDetail.numRegs) {
          this.manageReadWrite(u2a, {
            type:
              regModel._id === 'bq79600' ? 'readAllSingle600' : 'readAllSingle',
            device: deviceIndex,
            count: regReadCount,
            promiseArr: regReadPromiseArr
          });
        }

        return deferred.promise;
      }
    } else {
      regPage.isUpdating = true;
      console.log('SPI', 'Read', 'Regmap - Show progress', regPage);

      if (deviceIndex < 0) {
        readData.push(0xc0); // Broadcast read

        regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
        readData.push.apply(readData, regAddr); // Register address
        readData.push(0x00); // Read 1 byte of data
        crc = getByteArrFromHexStr(
          prependZero(getCRC16IBM(readData).toString(16), 4)
        );
        readData.push.apply(readData, crc);

        var deferred = Q.defer();

        var readPromise = Q.defer();
        readPromise.promise.then(function(data) {
          var readData = [];
          data = parseDeviceResponse(data);
          for (var i = 0; i < data.length; i++) {
            readData.push(data[i][4]);
          }
          readData = readData.reverse();
          deferred.resolve(readData);
          regPage.isUpdating = false;
          console.log(
            'SPI',
            'Read',
            'Broadcast read',
            'Regmap - Hide progress',
            regPage
          );
        });
        this.manageReadWrite(u2a, {
          type: 'read',
          data: readData,
          promise: readPromise
        });

        return deferred.promise;
      } else {
        readData.push(0x80); // Single device read intialization byte
        readData.push(deviceIndex); // Device address
        regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
        readData.push.apply(readData, regAddr); // Register address
        readData.push(0x00); // Read 1 byte of data
        crc = getByteArrFromHexStr(
          prependZero(getCRC16IBM(readData).toString(16), 4)
        );
        readData.push.apply(readData, crc); // CRC bytes

        var deferred = Q.defer();

        var readPromise = Q.defer();
        readPromise.promise.then(function(data) {
          deferred.resolve(data[4]);
          regPage.isUpdating = false;
          console.log(
            'SPI',
            'Read',
            'Single read',
            'Regmap - Hide progress',
            regPage
          );
        });
        this.manageReadWrite(u2a, {
          type: 'read',
          data: readData,
          promise: readPromise
        });

        return deferred.promise;
      }
    }
  };

  /**
   * Construct and send a command packet for writing to a register
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param info - The register map info for the register
   * @param value - The value to write to the register
   * @return - A promise that resolves immediately
   */
  CustomSPI_Interface.prototype.write = function(
    u2a,
    info,
    value,
    regModel,
    coreIndex
  ) {
    console.log('SPI', 'Write', u2a, info, value, coreIndex);
    var writeData = [];
    var regAddr, crc;

    var deviceIndex = getDeviceIndexSpi(coreIndex, regModel._id);

    if (deviceIndex > -1) {
      writeData.push(0x90); // Single device write intialization byte
      writeData.push(deviceIndex); // Device address
      regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
      writeData.push.apply(writeData, regAddr); // Register address
      writeData.push(value); // Data byte
      crc = getByteArrFromHexStr(
        prependZero(getCRC16IBM(writeData).toString(16), 4)
      );
      writeData.push.apply(writeData, crc); // CRC bytes

      console.log('SPI', 'Registermap', 'Write', writeData);
      this.manageReadWrite(u2a, {type: 'write', data: writeData});
    }

    return Q(true);
  };

  var readRegsHdlr = function(eventType, event) {
    console.log('SPI', 'Reg read event', eventType, event, event.detail);

    var detailObj = event.detail;
    if (detailObj.type === undefined) {
      detailObj.type = eventType;
    }

    var modelId = undefined;
    if (detailObj.regModelId) {
      modelId = detailObj.regModelId;
    }

    var deviceIndex = getDeviceIndexSpi(detailObj.coreIndex, modelId);

    console.log(
      'bc: read_regs: numRegs=' +
        detailObj.numRegs +
        ', regModel id=' +
        modelId +
        ', coreIndex=' +
        detailObj.coreIndex +
        ', type=' +
        detailObj.type
    );
    if (
      detailObj.type === 'init' ||
      detailObj.type === 'button' ||
      detailObj.type === 'refresh'
    ) {
      // setTimeout(function() {
      // var modelId = "undefined";

      var regPage = document.querySelector('ti-widget-register-page');
      regPage.isUpdating = true;
      console.log('SPI', 'Reg read event', 'Regmap - Show progress', regPage);

      if (regReadDetail && regReadDetail.modelId) {
        modelId = regReadDetail.modelId;
      }
      console.log(
        'regReadDetail.regModelId changed from ' +
          modelId +
          ' to ' +
          detailObj.regModelId
      );
      regReadDetail = detailObj;
      regReadDetail.device = deviceIndex;
      regReadCount = 0;
      regReadPromiseArr = [];
      // },0);
    }
  };

  var refreshPage = function() {
    var regPage = document.querySelector('ti-widget-register-page');
    regPage.readAllRegistersClickHdlr();
    templateObj.timerHdlr = window.setTimeout(
      refreshPage,
      templateObj.refreshInMs
    );
  };

  var logger = function(type, cmd) {
    document.querySelector('ti-bms-global').autoLogger(type, cmd);
  };

  var templateObj = document.querySelector('#template_obj');
  templateObj.timerHdlr = null;
  templateObj.refreshInMs = 100;

  // templateObj.addEventListener('dom-change', function(event) {
  if (document.querySelector('ti-bms-global')) {
    document.addEventListener(
      'refresh-regs',
      readRegsHdlr.bind(templateObj, 'refresh')
    );
    document.addEventListener(
      'init-regs',
      readRegsHdlr.bind(templateObj, 'init')
    );
    document.addEventListener(
      'read_regs',
      readRegsHdlr.bind(templateObj, 'statusOrButton')
    );
    document.addEventListener('read_reg', function(event) {
      regReadDetail = undefined;
    });
    document.addEventListener('auto_read_config_changed', function(event) {
      if (templateObj.timerHdlr) {
        window.clearTimeout(templateObj.timerHdlr);
        templateObj.timerHdlr = null;
      }
      var autoReadObj = event.detail.autoRead;
      if (autoReadObj.value > -1) {
        templateObj.refreshInMs = +autoReadObj.value;
        templateObj.timerHdlr = window.setTimeout(
          refreshPage,
          templateObj.refreshInMs
        );
      }
    });
  }
  // });
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
  init();
} else {
  document.addEventListener('WebComponentsReady', init);
}
